/*
 * Copyright (c) 2012, 2018 Oracle and/or its affiliates. All rights reserved.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0, which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the
 * Eclipse Public License v. 2.0 are satisfied: GNU General Public License,
 * version 2 with the GNU Classpath Exception, which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 */

package org.telegram.telegrambots.updatesreceivers;

import jakarta.annotation.Priority;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import jakarta.ws.rs.Priorities;
import jakarta.ws.rs.core.HttpHeaders;
import org.apache.commons.compress.compressors.CompressorException;
import org.apache.commons.compress.compressors.CompressorStreamFactory;
import org.glassfish.jersey.spi.ContentEncoder;

/**
 * Compression encoding support. Interceptor that encodes the output or decodes the input if
 * {@link HttpHeaders#CONTENT_ENCODING Content-Encoding header} value equals to most compression formats.
 *
 * @author Andrea Cavalli
 */
@Priority(Priorities.ENTITY_CODER)
public class CompressionEncoder extends ContentEncoder {

    private static final CompressorStreamFactory factory = new CompressorStreamFactory();

    /**
     * Initialize Encoder.
     */
    public CompressionEncoder() {
        super("deflate", "deflate64", "gzip", "zstd", "lz4", "lzma");
    }

    private String translateEncoding(String contentEncoding) throws IOException {
        switch (contentEncoding) {
            case "deflate":
                return CompressorStreamFactory.DEFLATE;
            case "deflate64":
                return CompressorStreamFactory.DEFLATE64;
            case "gzip":
                return CompressorStreamFactory.GZIP;
            case "zstd":
                return CompressorStreamFactory.ZSTANDARD;
            case "lz4":
                return CompressorStreamFactory.LZ4_BLOCK;
            case "lzma":
                return CompressorStreamFactory.LZMA;
            default:
                throw new IOException("Unsupported encoding " + contentEncoding);
        }
    }

    @Override
    public InputStream decode(String contentEncoding, InputStream encodedStream)
        throws IOException {
        try {
            String enc = translateEncoding(contentEncoding);
            return factory.createCompressorInputStream(enc, encodedStream);
        } catch (CompressorException e) {
            throw new IOException(e);
        }
    }

    @Override
    public OutputStream encode(String contentEncoding, OutputStream entityStream)
        throws IOException {
        try {
            String enc = translateEncoding(contentEncoding);
            return factory.createCompressorOutputStream(enc, entityStream);
        } catch (CompressorException e) {
            throw new IOException(e);
        }
    }
}
